# -*- coding: utf-8 -*-

from functools import wraps


class PipeLine(object):

    """
    A coroutine wrapper which enables pipelining syntax.

    :class:`PipeLine` allows you to flatten once-nested code just by wrapping
    your generators. The class provides combinators in the form of operators,
    allowing you to plug two generators together without having to nest lots of
    function calls. For example::

        >>> def summer(stdin):
        ...     sum = 0
        ...     for item in stdin:
        ...         sum += item
        ...         yield sum
        >>> pipeline = PipeLine(lambda: iter([1, 2, 3, 4])) | PipeLine(summer)
        >>> pipeline
        <PipeLine: <lambda> | summer>
        >>> for item in pipeline:
        ...     print item
        1
        3
        6
        10

    The yielded output of each generator in the chain becomes the input for the
    next. The rules for writing a pipeline function are simple:
    :class:`PipeLine` requires a callable which accepts a single argument (the
    input), and returns an iterator. The only exception is the first part of
    the pipeline, which should accept no arguments (as there will be no input).

    For convenience, two decorators have been provided in this module:
    :func:`source` and :func:`sink`. :func:`source` is for writing functions
    which only produce output, but take no input::

        >>> @source
        ... def my_generator():
        ...     yield 1
        ...     yield 2
        ...     yield 3
        >>> pl = my_generator()
        >>> pl
        <PipeLine: my_generator>
        >>> for item in pl:
        ...     print item
        1
        2
        3

    :func:`sink` is for writing functions which do accept input::

        >>> @sink
        ... def add_one(input):
        ...     for item in input:
        ...         yield item + 1
        >>> pl = my_generator() | add_one()
        >>> pl
        <PipeLine: my_generator | add_one>
        >>> for item in pl:
        ...     print item
        2
        3
        4

    With :func:`source` and :func:`sink`, your functions can also accept other
    parameters::

        >>> @sink
        ... def adder(input, amount):
        ...     for item in input:
        ...         yield item + amount
        >>> pl = my_generator() | adder(3)
        >>> pl
        <PipeLine: my_generator | adder>
        >>> for item in pl:
        ...     print item
        4
        5
        6
    """

    __slots__ = ('coro_func',)

    def __init__(self, coro_func):
        self.coro_func = coro_func

    @property
    def __name__(self):
        return self.coro_func.__name__

    def __repr__(self):
        return '<PipeLine: %s>' % getattr(self.coro_func, '__name__', repr(self.coro_func))

    def __or__(self, target):
        return target.__ror__(self)

    def __ror__(self, source):
        def pipe():
            return self.coro_func(iter(source))
        pipe.__name__ = '%s | %s' % (
                getattr(source, '__name__', repr(source)),
                getattr(self.coro_func, '__name__', repr(self.coro_func)))
        return PipeLine(pipe)

    def __iter__(self):
        return self.coro_func()


def source(func):
    """
    Wrap a function as a pipeline source (i.e. one that does not accept stdin).

        >>> @source
        ... def echoer(items):
        ...     for item in items:
        ...         yield item
        >>> e = echoer([1, 2, 3])
        >>> e
        <PipeLine: echoer>
        >>> for item in e:
        ...     print item
        1
        2
        3
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        coro_func = wraps(func)(lambda: func(*args, **kwargs))
        return PipeLine(coro_func)
    return wrapper


def sink(func):
    """
    Wrap a function as a pipeline sink (i.e. one that accepts stdin).

        >>> @sink
        ... def printer(stdin, outfile=None):
        ...     for item in stdin:
        ...         print >>outfile, item
        ...         yield item
        >>> p = printer()
        >>> p
        <PipeLine: printer>
        >>> output = list(iter([1, 2, 3, 4, 5]) | p)
        1
        2
        3
        4
        5
        >>> output
        [1, 2, 3, 4, 5]
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        @wraps(func)
        def coro_func(stdin=()):
            return func(stdin, *args, **kwargs)
        return PipeLine(coro_func)
    return wrapper



# -*- coding: utf-8 -*-

import re

from calabash.pipeline import source, sink


@source
def echo(item):
    """
    Yield a single item. Equivalent to ``iter([item])``, but nicer-looking.

        >>> list(echo(1))
        [1]
        >>> list(echo('hello'))
        ['hello']
    """
    yield item

@source
def cat(*args, **kwargs):
    r"""
    Read a file. Passes directly through to a call to `open()`.

        >>> src_file = __file__.replace('.pyc', '.py')
        >>> for line in cat(src_file):
        ...     if line.startswith('def cat'):
        ...          print repr(line)
        'def cat(*args, **kwargs):\n'
    """
    return iter(open(*args, **kwargs))


@source
def curl(url):
    """
    Fetch a URL, yielding output line-by-line.

        >>> UNLICENSE = 'http://unlicense.org/UNLICENSE'
        >>> for line in curl(UNLICENSE):
        ...     print line, # doctest: +ELLIPSIS
        This is free and unencumbered software released into the public domain.
        ...
    """
    import urllib2
    conn = urllib2.urlopen(url)
    try:
        line = conn.readline()
        while line:
            yield line
            line = conn.readline()
    finally:
        conn.close()


@sink
def grep(stdin, pattern_src):
    """
    Filter strings on stdin for the given regex (uses :func:`re.search`).

        >>> list(iter(['cat', 'cabbage', 'conundrum', 'cathedral']) | grep(r'^ca'))
        ['cat', 'cabbage', 'cathedral']
    """
    pattern = re.compile(pattern_src)
    for line in stdin:
        if pattern.search(line):
            yield line


@sink
def sed(stdin, pattern_src, replacement, exclusive=False):
    """
    Apply :func:`re.sub` to each line on stdin with the given pattern/repl.

        >>> list(iter(['cat', 'cabbage']) | sed(r'^ca', 'fu'))
        ['fut', 'fubbage']

    Upon encountering a non-matching line of input, :func:`sed` will pass it
    through as-is. If you want to change this behaviour to only yield lines
    which match the given pattern, pass `exclusive=True`::

        >>> list(iter(['cat', 'nomatch']) | sed(r'^ca', 'fu'))
        ['fut', 'nomatch']
        >>> list(iter(['cat', 'nomatch']) | sed(r'^ca', 'fu', exclusive=True))
        ['fut']
    """
    pattern = re.compile(pattern_src)
    for line in stdin:
        match = pattern.search(line)
        if match:
            yield (line[:match.start()] +
                   match.expand(replacement) +
                   line[match.end():])
        elif not exclusive:
            yield line


@sink
def pretty_printer(stdin, **kwargs):
    """
    Pretty print each item on stdin and pass it straight through.

        >>> for item in iter([{'a': 1}, ['b', 'c', 3]]) | pretty_printer():
        ...     pass
        {'a': 1}
        ['b', 'c', 3]
    """
    import pprint
    for item in stdin:
        pprint.pprint(item, **kwargs)
        yield item


@sink
def map(stdin, func):
    """
    Map each item on stdin through the given function.

        >>> list(xrange(5) | map(lambda x: x + 2))
        [2, 3, 4, 5, 6]
    """
    for item in stdin:
        yield func(item)


@sink
def filter(stdin, predicate):
    """
    Only pass through items for which `predicate(item)` is truthy.

        >>> list(xrange(5) | filter(lambda x: x % 2 == 0))
        [0, 2, 4]
    """
    for item in stdin:
        if predicate(item):
            yield item


@sink
def sh(stdin, command, check_success=False):
    r"""
    Run a shell command, send it input, and produce its output.

        >>> print ''.join(echo("h\ne\nl\nl\no") | sh('sort -u'))
        e
        h
        l
        o
        <BLANKLINE>
        >>> for line in sh('echo Hello World'):
        ...     print line,
        Hello World
        >>> for line in sh('false', check_success=True):
        ...     print line, # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        CalledProcessError: Command '['false']' returned non-zero exit status 1
    """
    import subprocess
    import shlex

    if isinstance(command, basestring):
        command = shlex.split(command)

    pipe = subprocess.Popen(command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE)

    try:
        for line in stdin:
            pipe.stdin.write(line)
        pipe.stdin.close()
        for line in pipe.stdout:
            yield line
    finally:
        result = pipe.wait()
        if check_success and result != 0:
            raise subprocess.CalledProcessError(result, command)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path as p

from distribute_setup import use_setuptools; use_setuptools()
from setuptools import setup, find_packages

VERSION = open(p.join(p.dirname(p.abspath(__file__)), 'VERSION')).read().strip()

setup(
    name             = 'calabash',
    version          = VERSION,
    author           = "Zachary Voase",
    author_email     = "z@zacharyvoase.com",
    url              = 'http://github.com/zacharyvoase/calabash',
    description      = "Bash-style pipelining syntax for Python generators.",
    packages         = find_packages(where='src'),
    package_dir      = {'': 'src'},
    test_suite       = 'calabash._get_tests',
)
